#!/usr/bin/env python
"""
    confluence output
    
    Copyright (C) 2012 Pablo Martin <pablo at odkq.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import cgi
import base64
import tempfile
import zerodoc.diagram
import zerodoc.utils

def write_listlines(proc, options, listlines, toc):
    lastlevel = 0
    o = ''
    for uline in listlines:
        s = '*'
        for i in range(uline['level']):
            s += '*'
        if toc:
            s += ' [' + proc.process(uline) + '|#' +\
                zerodoc.utils.get_anchor(uline['string']) + ']\n'
        else:
            s += ' ' + proc.process(uline) + '\n'
        o += s
    return o

def write_sourcelines(proc, options, sourcelines):
    o = '{code:none}\n'
    for sline in sourcelines:
        if 'string' in sline:
            o += proc.process(sline) + '\n'
        else:
            print 'unknown element in sourcelines ' + str(sline) + ' ' + str(sline.keys())
    o += '{code}\n'
    return o

def write_textlines(proc, options, textlines):
    o = '\n'
    for t in textlines:
        o += proc.process(t) + ' '
    o += '\n'
    return o

def write_diagramlines(proc, options, diagramlines):
    dlines = []
    if 'rawdiagrams' in options:
        return write_sourcelines(options, diagramlines)
    for sline in diagramlines:
        if 'string' in sline:
            dlines.append(proc.process(sline))
        else:
            print 'unknown element in sourcelines (diagram)' + str(sline) +\
            ' ' + str(sline.keys())
    img = zerodoc.diagram.get_diagram_ditaa(options, dlines)
    if img == None:
        return ''
    if not 'datauri' in options:
        if not os.path.exists('images'):
            os.mkdir('images')
        f = tempfile.NamedTemporaryFile(delete=False, prefix='zero', dir='images')
        f.write(img)
        n = f.name
        f.close()
        return '{html}\n<img src="' + n + '" />\n{html}\n'

    duri = str(base64.encodestring(img)).replace("\n", "")
    if not 'svg' in options:
        t = '{html}' + '\n<img alt="sample" src="data:image/png;base64,{0}">\n'.format(duri) + '{html}\n'
    else:
        t = '{html}' + '\n<img alt="sample" src="data:image/svg;base64,{0}">\n'.format(duri) + '{html}\n'
    return t

 
def write_confluence_paragraph(proc, options, para, toc = False):
    o = ''
    if para.has_key('sourcelines'):
        o += write_sourcelines(proc, options, para['sourcelines'])
    elif para.has_key('listlines'):
        o += write_listlines(proc, options, para['listlines'], toc)
    elif para.has_key('textlines'):
        o += write_textlines(proc, options, para['textlines'])
    elif para.has_key('diagramlines'):
        o += write_diagramlines(proc, options, para['diagramlines'])
    return o

def write_confluence_section(proc, options, section):
    o = '\n'
    o += '{anchor:' + zerodoc.utils.get_anchor(section['title']['string']) +'}\n'
    o += 'h' + str(section['level'] + 1) + '. '
    o += section['title']['string'] + '\n'
    for para in section['paragraphs']:
        o += write_confluence_paragraph(proc, options, para)
    return o

def write(doc, options = ['ditaa', 'datauri']):
    '''
    Output the confluence wiki rendering of a doc tree
    '''
    proc = zerodoc.utils.Processor(doc, '[{0}|{1}]')
    o = ''
    # todo put title in post
    o += 'h1. '
    for titleline in doc['header']['title']['textlines']:
        o += proc.process(titleline) + ' '
    o += '\n'
    for para in doc['header']['abstract']['abstract']:
       o += write_confluence_paragraph(proc, options, para)
    # o += '<h2>Table of contents</h2>'
    o += write_confluence_paragraph(proc, options, doc['header']['toc'], toc=True)
    for section in doc['body']['sections']:
        o += write_confluence_section(proc, options, section) 
    return o




#!/usr/bin/env python
""" 
    zerodoc reStructuredText (http://docutils.sourceforge.net/rst.html)
    output
    
    Copyright (C) 2014 Pablo Martin <pablo at odkq.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import os.path
import cgi
import base64
import string
import tempfile
import subprocess
import zerodoc.diagram
import zerodoc.utils
import zerodoc.html     # For raw html pass-through
# (http://docutils.sourceforge.net/docs/ref/rst/directives.html\
#  raw-data-pass-through)
import rsvg

def write_listlines(proc, options, listlines, toc):
    lastlevel = 0
    # o = '<ul>\n'
    o = '\n'
    for uline in listlines:
        # We can go from 3 to 0, but not to 1 to 3 (an
        # intermediate title is needed)
        # TODO: Do this test in the parser
        d = lastlevel - uline['level']
        if d == -1:
            # Certain formats make list delimiters depend on level, - * +
            # and so on. It's not the case of HTML
            lastlevel += 1
        elif d > 0:
            # Unwind <ul's>
            for i in range(d):
                # o += '</ul>\n'
                lastlevel -= 1
        elif uline['level'] != lastlevel:
            # List indentation error
            # TODO: Notify from the parser?
            o += '<code>error, new listlevel '
            o += str(uline['level'])
            o += ' oldlistlevel ' + str(lastlevel)
            o += '</code>'
        if toc:
            o += ' ' * (uline['level'] * 2) + '- '
            # o += '<a href="#' + zerodoc.utils.get_anchor(uline['string']) +'">'
            o += proc.process(uline) + '\n\n'
            # o += '</a>\n\n'
        else:
            o += ' ' * (uline['level'] * 2) + '- '
            o += proc.process(uline) + '\n\n'
    # Unwind last <ul's>
    # for i in range(lastlevel):
    #   o += '</ul>\n'
    #o += '</ul>\n'
    return o

def write_sourcelines(proc, options, sourcelines):
    o = '::\n\n'
    for sline in sourcelines:
        if 'string' in sline:
            o += proc.process(sline) + '\n'
        else:
            print 'no string in sourceline? ' + str(sline) + ' ' + str(sline.keys())
    return o

def write_deflist(proc, options, deflist):
    o = '\n'        # Indented lines need a blank line before in rest??
    for t in deflist['term']['textlines']:
        o += proc.process(t) + '\n'
    for t in deflist['definition']['textlines']:
        # There is no need to overindent, as restructuredText indents text
        # automatically
        o += proc.process(t) + '\n'
    return o

def write_textlines(proc, options, textlines):
    o = '\n'
    for t in textlines:
        o += proc.process(t) + '\n'
    return o

def write_diagramlines(proc, options, diagramlines):
    # o = '\n\n.. raw:: html\n\n'
    # o += zerodoc.html.write_diagramlines(proc, options, diagramlines)
    dlines = []
    if 'rawdiagrams' in options:
        return write_sourcelines(options, diagramlines)
    for sline in diagramlines:
        if 'string' in sline:
            dlines.append(sline['string'])
        else:
            print 'unknown element in sourcelines (diagram)' + str(sline) +\
            ' ' + str(sline.keys())
    img, ext = zerodoc.diagram.get_diagram(options, dlines)
    if img == None:
        return ''
    if not 'datauri' in options:
        if not os.path.exists('images'):
            os.mkdir('images')
        f = tempfile.NamedTemporaryFile(suffix='.' + ext, delete=False, prefix='zero', dir='images')
        f.write(img)
        n = f.name
        f.close()
        return '\n.. image:: images/' + os.path.basename(n) + '\n\n'
    # return o

 
def write_rst_paragraph(proc, options, para, toc = False):
    o = ''
    if para.has_key('sourcelines'):
        o += write_sourcelines(proc, options, para['sourcelines'])
    elif para.has_key('listlines'):
        o += write_listlines(proc, options, para['listlines'], toc)
    elif para.has_key('textlines'):
        o += write_textlines(proc, options, para['textlines'])
    elif para.has_key('diagramlines'):
        o += write_diagramlines(proc, options, para['diagramlines'])
    elif para.has_key('deflist'):
        o += write_deflist(proc, options, para['deflist'])
    return o

def write_rst_section(proc, options, section):
    o = ''
    # o += '<a name="' + zerodoc.utils.get_anchor(section['title']['string']) + '"></a>'
    # if section['level'] == 
    o += section['title']['string'] + '\n'
    subtitles = ['=', '-', '~']
    if section['level'] > 2:
        subtitle = '~'
    else:
        subtitle = subtitles[section['level']]
    o += subtitle * len(section['title']['string']) + '\n'
    for para in section['paragraphs']:
        o += write_rst_paragraph(proc, options, para)
    o += '\n\n'
    return o

def write(doc, options = ['ditaa', 'datauri', 'notoc']):
    '''
    Output the reStructuredText rendering of a doc tree
    '''
    proc = zerodoc.utils.Processor(doc, '`{0} <{1}>`_')
    o = ''
    # o += '<h1>\n'
    rst_title = ''
    for titleline in doc['header']['title']['textlines']:
        rst_title += proc.process(titleline) # + '\n'
    o += '=' * len(rst_title) + '\n'
    o += rst_title + '\n'
    o += '=' * len(rst_title) + '\n'
    # o += '</h1>\n'
    # o += '<h2>Abstract</h2>'
    for para in doc['header']['abstract']['abstract']:
        if para['textlines'][0]['string'] == 'Table of contents':
            continue
        o += write_rst_paragraph(proc, options, para)
    o += '\n'
    if not 'notoc' in options:
        # Usually reStructuredText processors attach their own index in the
        # side (sphinx, for example). In that case, you better do not output
        # the toc (it is still used)
        o += 'Table of contents\n'
        o += '-----------------\n'
        o += write_rst_paragraph(proc, options, doc['header']['toc'], toc=True)
    for section in doc['body']['sections']:
        o += write_rst_section(proc, options, section) 
    return o




#!/usr/bin/env python
""" 
    zerodoc ascii-art diagrams support
    
    Copyright (C) 2012 Pablo Martin <pablo at odkq.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import cgi
import rsvg
import string
import hashlib
import tempfile
import subprocess

def generate_diagram(buffer, format):
    # Write the buffer into a file
    f = tempfile.NamedTemporaryFile(delete=False, prefix='zero')
    f.write(buffer)
    f.close()
    os.de

def extract_option(options, key):
    for o in options:
        b = o.split(':')
        if len(b) == 2:
            if b[0] == key:
                return b[1]
    return None

def generate_diagram_ditaa(path, options):
    jarpath = extract_option(options, 'jarpath')
    if jarpath == None:
        r = ['ditaa', path ]
    else:
        r = ['java', '-jar', jarpath, path ]
    p = subprocess.Popen(r, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    o = p.communicate()[0]
    i = o.find('Rendering to file: ')
    if i == -1:
        print 'Error generating diagram!'
        return None
    j = o[i:].find('\n')
    return o[i+19:i+j], 'png'

def generate_diagram_aafigure(path, options):
    # aafigure diagram -t svg -o diagram.svg
    if 'svg' in options:
        ext = 'svg'
        outpath = path + '.svg'
        r = ['aafigure', path, '-t', 'svg', '-o', outpath]
    else:
        ext = 'png'
        outpath = path + '.png'
        r = ['aafigure', path, '-t', 'png', '-o', outpath]
    p = subprocess.Popen(r, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    o = p.communicate()[0]
    return outpath, ext

def generate_diagram_a2s(path, options):
    # aafigure diagram -t svg -o diagram.svg
    outpath = path + '.svg'
    r = ['php5', '/usr/share/asciitosvg/a2s' ]
    # , path, '-t', 'svg', '-o', outpath]
    with open(os.devnull, "w") as fnull:
        p = subprocess.Popen(r, stdout=open(outpath, 'w'),
            stdin=open(path, 'r'), stderr = fnull)
    p.communicate()
    return outpath, 'svg'

def generate_diagram_tikz(path, options):
    ''' generate_diagram_tikz '''
    texfile = path + '.tex'
    pdffile = path + '.pdf'
    psfile = path + '.eps'
    pngfile = path + '.png'
    auxfile = path + '.aux'
    logfile = path + '.log'

    t = open(texfile, 'wb')
    s = open(path, 'r')
    t.write('\\documentclass{article}\n')
    t.write('\\usepackage[pdftex,active,tightpage]{preview}\n')
    if 'border' in options:
        t.write('\\setlength\PreviewBorder{2mm}\n')
    t.write('\\usepackage{tikz}\n')
    t.write('\\begin{document}\n')
    t.write('\\begin{preview}\n')
    t.write(s.read())
    t.write('\\end{preview}\n\\end{document}\n')
    t.flush()
    os.fsync(t.fileno())
    t.close()
    s.close()

    dpis = extract_option(options,'dpis')
    if dpis == None:
        dpis = '150'
    last_cwd = os.getcwd()
    os.chdir('/tmp')
    try:
        subprocess.check_call(['pdflatex', texfile ], stdin=subprocess.PIPE,
                stdout=subprocess.PIPE)
        subprocess.check_call(['pdftops', '-eps', pdffile ])
        subprocess.check_call(['convert', '-density', dpis, psfile, pngfile ])
    except CalledProcessError:
        print 'Error converting to tikz somehow'
        pngfile = None
    for file in [ pdffile, psfile, auxfile, logfile, texfile ]:
        if os.path.exists(file):
            os.remove(file)
    os.chdir(last_cwd)
    return pngfile

def generate_diagram_gnuplot(path, options):
    return None

def detect_first_line(path):
    f = open(path, 'r')
    t = f.readlines()
    for fmt in [ 'ditaa', 'aafigure', 'asciitosvg' ]:
        if t[0].find(fmt) != -1:
            f = tempfile.NamedTemporaryFile(delete=False)
            f.write(string.join(t[1:]))
            n = f.name
            f.close()
            return n, fmt
    return path, None

def detect_diagram_type(path, options):
    # The diagram can be explicitly tagged on
    # it's first line. Otherwise it is deducted
    # from it's content or the default option passed
    # on the command line
    newpath, t = detect_first_line(path)
    if t != None:
        return newpath, t

    types = {
        'tikz' : [ '\\begin{tikzpicture}' ],
        'gnuplot' : ['plot']
    }
    f = open(path, 'r')
    lines = f.readlines()
    for l in lines:
        for k in types.keys():
            for kw in types[k]:
                if l.find(kw) != -1:
                    return path, k
    for fmt in [ 'ditaa', 'aafigure', 'asciitosvg' ]:
        if fmt in options:
            return path, fmt
    return path, 'unknown'

# get_diagram:
# Return the diagram bitmap in a buffer
def get_diagram(options, lines):
    f = tempfile.NamedTemporaryFile(delete=False)
    for line in lines:
        f.write(line + '\n')
    f.close()
    p = f.name
    n, t = detect_diagram_type(p, options)
    # if type == 'unknown':
    if 'ditaa' == t:
        dfile, ext = generate_diagram_ditaa(n, options)
    elif 'aafigure' == t:
        dfile, ext = generate_diagram_aafigure(n, options)
    elif 'asciitosvg' == t:
        dfile, ext = generate_diagram_a2s(n, options)
    elif t == 'tikz':
        dfile = generate_diagram_tikz(n, options)
        ext = 'png'
    elif t == 'gnuplot':
        dfile = generate_diagram_gnuplot(n, options)
    else:
        print 'Specify a default conversor for diagrams! (ditaa/aafigure/a2s)'
        return None, None
    if dfile == None:
        return None
    f = open(dfile, 'r')
    r = f.read()
    if 'svg' in options and 'aafigure' == t:
        # Set width on line four of the outputted svg
        handle = rsvg.Handle(file=dfile)
        lines = r.split('\n')
        lines[4] += (' width="' + str(handle.props.width) +
                     'px" height="' + str(handle.props.height) + 'px"')
        r = string.join(lines, '\n')
    f.close()
    # os.remove(n)
    return r, ext




#!/usr/bin/env python
""" 
    zerodoc html 2.0 (http://tools.ietf.org/rfc/rfc1866.txt) output
    
    Copyright (C) 2012 Pablo Martin <pablo at odkq.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import cgi
import base64
import string
import tempfile
import subprocess
import zerodoc.diagram
import zerodoc.utils
import rsvg

def write_listlines(proc, options, listlines, toc):
    lastlevel = 0
    o = '<ul>\n'
    for uline in listlines:
        # We can go from 3 to 0, but not to 1 to 3 (an
        # intermediate title is needed)
        # TODO: Do this test in the parser
        d = lastlevel - uline['level']
        if d == -1:
            # Certain formats make list delimiters depend on level, - * +
            # and so on. It's not the case of HTML
            o += '<ul>\n'
            lastlevel += 1
        elif d > 0:
            # Unwind <ul's>
            for i in range(d):
                o += '</ul>\n'
                lastlevel -= 1
        elif uline['level'] != lastlevel:
            # List indentation error
            # TODO: Notify from the parser?
            o += '<code>error, new listlevel '
            o += str(uline['level'])
            o += ' oldlistlevel ' + str(lastlevel)
            o += '</code>'
        if toc:
            o += '<li>'
            o += '<a href="#' + zerodoc.utils.get_anchor(uline['string']) +'">'
            o += proc.process(uline)
            o += '</a></li>\n'
        else:
            o += '<li>' 
            o += proc.process(uline)
            o += '</li>\n'
    # Unwind last <ul's>
    for i in range(lastlevel):
        o += '</ul>\n'
    o += '</ul>\n'
    return o

def write_sourcelines(proc, options, sourcelines):
    if 'pygments' in options:
        o = '<pre>\n'
        for sline in sourcelines:
            if 'string' in sline:
                o += proc.process(sline) + '\n'
            else:
                print 'no string in sourceline? ' + str(sline) + ' ' + str(sline.keys())
        o += '</pre>\n'
    else:
        f = tempfile.NamedTemporaryFile(delete=False,
            prefix='zero', dir='/tmp')
        for sline in sourcelines:
            if 'string' in sline:
                # Do not process line
                f.write(sline['string'] + '\n')
                # f.write(proc.process(sline) + '\n')
        n = f.name
        f.close()
        output = n + '.html'
        subprocess.call(['pygmentize', '-f', 'html', '-l', 'c', '-o',
            output, n])
        o = open(output).read()
        os.remove(n)
        os.remove(output)
    return o

def write_deflist(proc, options, deflist):
    o = '<dl>'
    o += '<dt>'
    for t in deflist['term']['textlines']:
        o += proc.process(t) + '\n'
    o += '<dd>'
    for t in deflist['definition']['textlines']:
        o += proc.process(t) + '\n'
    o += '</dl>'
    return o

def write_textlines(proc, options, textlines):
    o = '<p>\n'
    for t in textlines:
        o += proc.process(t) + '\n'
    o += '</p>\n'
    return o

def write_diagramlines(proc, options, diagramlines):
    dlines = []
    if 'rawdiagrams' in options:
        return write_sourcelines(options, diagramlines)
    for sline in diagramlines:
        if 'string' in sline:
            dlines.append(sline['string'])
        else:
            print 'unknown element in sourcelines (diagram)' + str(sline) +\
            ' ' + str(sline.keys())
    img, ext = zerodoc.diagram.get_diagram(options, dlines)
    if img == None:
        return ''
    if not 'datauri' in options:
        if not os.path.exists('images'):
            os.mkdir('images')
        f = tempfile.NamedTemporaryFile(suffix='.' + ext, delete=False, prefix='zero', dir='images')
        f.write(img)
        n = f.name
        f.close()
        return '<img src="' + n + '" />\n'

    if ext == 'png':
        # Embed png using 'datauri' old trick
        duri = str(base64.encodestring(img)).replace("\n", "")
        t = '<img alt="sample" src="data:image/' + ext + ';base64,{0}">\n'.format(duri)
    else:
        # Embed svg directly in the html file
        # remove first 4 lines (<xml> and doctype) and set witdth and height
        # attributes that aafigure does not set)
        lines = img.split('\n')[4:]
        t = string.join(lines,'\n')
        
    return t

 
def write_html_paragraph(proc, options, para, toc = False):
    o = ''
    if para.has_key('sourcelines'):
        o += write_sourcelines(proc, options, para['sourcelines'])
    elif para.has_key('listlines'):
        o += write_listlines(proc, options, para['listlines'], toc)
    elif para.has_key('textlines'):
        o += write_textlines(proc, options, para['textlines'])
    elif para.has_key('diagramlines'):
        o += write_diagramlines(proc, options, para['diagramlines'])
    elif para.has_key('deflist'):
        o += write_deflist(proc, options, para['deflist'])
    return o

def write_html_section(proc, options, section):
    o = ''
    o += '<a name="' + zerodoc.utils.get_anchor(section['title']['string']) + '"></a>'
    o += '<h' + str(section['level'] + 2) 
    o += '>' + section['title']['string'] + '</h'
    o += str(section['level'] + 2) + '>'
    for para in section['paragraphs']:
        o += write_html_paragraph(proc, options, para)
    return o

def write(doc, options = ['ditaa', 'datauri']):
    '''
    Output the HTML 2.0 rendering of a doc tree
    '''
    proc = zerodoc.utils.Processor(doc, '<a href="{1}">{0}</a>', cgi.escape)
    o = ''
    if not 'noheaders' in options:
        if not 'html5' in options:
            # i know, i know..
            o += '<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0 Strict//EN">\n'
        else:
            o += '<!DOCTYPE html>\n'
        o += '<html>\n'
        o += '<head>\n'
        if not 'html5' in options:
            o += '<meta http-equiv="Content-Type" content="text/html;charset=iso-8859-1" >'
        else:
            o += '<meta charset="utf-8">\n'
            css = None
            for option in options:
                if option[:3] == 'css':
                    css = option.split(':')[1]
            if css:
                if 'datauri' in options:
                    o += '<style media="screen" type="text/css">'
                    o += open(css).read()
                    o += '</style>'
                else:
                    o += '<link rel="stylesheet" href="' + css + '">\n'
        o += '<title>\n'
        for titleline in doc['header']['title']['textlines']:
            o += proc.process(titleline) + '\n'
        o += '</title>\n'
        o += '<body>\n'
    o += '<h1>\n'
    for titleline in doc['header']['title']['textlines']:
        o += proc.process(titleline) + '\n'
    o += '</h1>\n'
    # o += '<h2>Abstract</h2>'
    for para in doc['header']['abstract']['abstract']:
        if para['textlines'][0]['string'] == 'Table of contents':
            continue
        o += write_html_paragraph(proc, options, para)
    o += '<h3>Table of contents</h3>'
    o += write_html_paragraph(proc, options, doc['header']['toc'], toc=True)
    for section in doc['body']['sections']:
        o += write_html_section(proc, options, section) 
    if not 'noheaders' in options:
        o += '</body>'
        o += '</html>'
    return o




#!/usr/bin/env python
""" 
    Convenient functions that can be used by various renderers
 
    Copyright (C) 2012 Pablo Martin <pablo at odkq.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import hashlib
import cgi

def get_anchor(s):
    '''
    Some standard way to extract and refer to internal hyperlins 
    (called anchors for some markups)
    '''
    h = hashlib.new('ripemd160')
    h.update(s.lower())
    return h.hexdigest()[:8]

class Processor():
    '''
    Include attributes and links in a line with
    a format depending in the output format

    '''
    def __init__(self, doc, format_string, escape_function = None):
        '''
        escape_function is a function that will be called for the
        text parts of the output (cgi.escape for html output for
        example) Pass None to include those verbatim
        '''
        if not escape_function:
            self.escape_function = lambda x: x
        else:
            self.escape_function = escape_function
        self.format_string = format_string
        self.doc = doc

    def process(self, line):
        s = ''
        a = line['string']
        l = 0
        r = []
        if 'references' in line:
            for reference in line['references']:
                j = reference[1] - l
                if j > 0:
                    s += self.escape_function(a[:j])
                try:
                    s += self.format_string.format(reference[0], self.doc['links'][reference[0]])
                except KeyError:
                    return self.escape_function(line['string'])
                a = a[j:]
                l = reference[1]
            if a != '':
                s += self.escape_function(a)
            return s
        else:
            return self.escape_function(a)




#!/usr/bin/env python
from distutils.core import setup

classifiers = [
   "Development Status :: 3 - Alpha",
   "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
   "Programming Language :: Python :: 2.6",
   "Topic :: Text Processing :: Markup",
   "Topic :: Text Processing :: Filters",
   "Topic :: Software Development :: Documentation",
   "Environment :: Console",
]

setup(
    name = 'zerodoc',
    version = '0.0.1',
    description = "minimalistic {asciidoc/pod/phpdoc}-alike " +
                  "plaintext to html/markdown/whatever' format" +
                  "with a very simple interpreter and a very" +
                  "simple syntax",
    author = "Pablo Martin",
    author_email = "pablo@odkq.com",
    packages = ['zerodoc'],
    scripts = [
        'scripts/zerodoc'
    ],
    url = "https://github.com/odkq/zerodoc",
    license = "GPL v3",
    # README.0 is written in zerodoc itself
    long_description = open('README.rst').read(),
    data_files = [
                ("/usr/share/doc/zerodoc",
                        [ "README.zd"]),
                ("/usr/share/zerodoc/stylesheets",
                        [ "stylesheets/default.css"]),
    ],
    classifiers = classifiers
)




#!/usr/bin/env python
""" 
    zerodoc

    Very simple interpreter for a 'plain text format' named zerodoc
    that is parsed into a tree and from there to HTML 2.0/markdown etc

    Built upon the PLY 'lex & yacc for python'. The yacc files and the
    lex regular expressions defining zerodoc can be found interspersed
    in the t_ and p_ functions of this file

    Copyright (C) 2012 Pablo Martin <pablo at odkq.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import sys
import ply.lex as lex
import ply.yacc as yacc

tokens = ( 'SOURCE', 'TEXT', 'TEXT2', 'TEXTLIST', 'FIRSTLIST', 'FIRSTSOURCE' ,
        'FIRSTDIAGRAM', 'NEWLINE')

t_TEXT=r'[^-\n\ ][^\n]+'
t_TEXT2=r'[^\n\ ][^\ \n][^\n]+'
t_TEXTLIST=r'[\ ]+-[^\n]+'
t_FIRSTLIST=r'-[^\n]+'
t_SOURCE=r'[^-\n][^\n]*'
t_FIRSTSOURCE=r'\ [^-\n\ ][^\n]*'
t_FIRSTDIAGRAM=r'\ [\ ]+[^-\n\ ][^\n]*'

start = 'document'

# TOC, or not
in_header = True

# Link to the base of the tree after parsing (yacc.yacc()
def t_NEWLINE(t):
    r'\n'
    t.lexer.lineno += len(t.value)
    return t

def t_error(t):
    print "Illegal character '%s'" % t.value[0]
    return False

def test(lexer,data):
    lexer.input(data)
    while 1:
         tok = lexer.token()
         if not tok: break
         print tok

def append_or_create(name, p):
    ''' When working on a rule that 'groups' several
        elements together, generate an array if it is
        the first element or append to the existing one'''
    if len(p) == 3:     # elements: elements element
        p[0] = p[1]
        p[0][name].append(p[2])
    else:               # elements: element (first occurrence)
        p[0] = { name: [] }
        p[0][name].append(p[1])
    return p[0]

# Abstract can only have text paragrahps (no lists
# or sourcelines
def p_textparas(p):
    '''textparas : textparas textpara
                | textpara'''
    append_or_create('abstract', p)

def p_paragraphs(p):
    '''paragraphs : paragraphs paragraph
                  | paragraph'''
    append_or_create('paragraphs', p)

def in_toclist(s, d):
    found = None
    level = None
    toc_list = d['header']['toc']
    for e in toc_list['listlines']:
        if e['string'].lower() == s.lower():
            if found != None:
                print 'ERROR: The element ' + e['string'] + 'in TOC is duplicated!'
                sys.exit(1)
            else:
                found = e['string']
                level = e['level']
    return found, level

def p_paragraph(p):
    '''paragraph : textlines NEWLINE
                 | sourcelines NEWLINE
                 | listlines NEWLINE
                 | diagramlines NEWLINE
                 | deflist NEWLINE
    '''
    # A paragraph made from listlines can be a list of
    # links
    if 'listlines' in p[1]:
        pass
        # print 'p_paragraph(): [' + str(p[1]['listlines']) + ']'
    p[0] = p[1]

def extract_attributes(x):
    """
    Extract attributes (bold, underline, quote) from text
    >>> extract_attributes("this is *bold*, this is _underline_,\
 this 'quote'")
    {'modified_string': 'this is bold, this is underline, this quote',\
 'bold':[[8,11]], 'underline':[[23,31]], 'quote': [[39,43]]}
    """
    pass

def extract_links(x):
    """Extract links/references from a line. Returns a dictionary
    of {'links':{'keyword': link,...,'keyword': link},
    'references':[[keyword,pos],...,[keyword,pos]],
    'modified_string': s }

    The modified string will not have nor links or references
    on it, and the positions in the array of references
    for are relative to this modified string, not to the
    original one

    >>> extract_links('text line `name`:http://pepe.com continue')
    {'modified_string': 'text line  continue', 'references': [['name', 10]],\
 'links': {'name': 'http://pepe.com'}}

    >>> extract_links('text line `reference1` and `reference2`')
    {'modified_string': 'text line  and ', 'references':\
 [['reference1', 10], ['reference2', 15]], 'links': {}}

    >>> extract_links('text line with no links whatsoever')
    {'modified_string': 'text line with no links whatsoever',\
 'references': [], 'links': {}}

    """
    modified_string = ''
    offset = 0
    references = []
    links = {}
    while True:
        s = x[offset:]
        local_modified_string = ''
        first = s.find('`')
        if first == -1:
            modified_string += s
            break
        second = s[first+1:].find('`')
        if second == -1:
            modified_string += s
            break
        if (second+first+2) >= len(s) or s[second+first+2] != ':':
            # no link afterwards, thus it is a reference
            pos = first + len(modified_string)
            local_modified_string += s[:first]
            # local_modified_string += s[(second+first+3):]
            offset += second+first+2
            ref = s[first+1:second+first+1]
            references.append([ref, pos])
        else:
            third = s[second+first+3:].find(' ')
            if third == -1:
                # eol
                third = len(s)
            pos = first + len(modified_string)
            local_modified_string += s[:first]
            # local_modified_string += s[second+first+third+3:]
            offset += second+first+third+3
            ref = s[first+1:second+first+1]
            link = s[second+first+3:second+first+third+3]
            references.append([ref, pos])
            links[ref] = link
        modified_string += local_modified_string
    return {'links': links, 'references':references,
            'modified_string': modified_string }

def p_textpara(p):
    '''textpara : textlines NEWLINE'''
    p[0] = p[1]

# Lists with only 1 element are not allowed
def p_listlines(p):
    '''listlines : firstlist listline
                 | listlines listline
                 | listlines firstlist
                 | firstlist firstlist'''
    if 'listlines' in p[1]:
        p[0] = p[1] 
        p[0]['listlines'].append(p[2])
    else:
        p[0] = { 'listlines': [p[1], p[2]] }

# Used both for diagrams and source
def insert_source_diagram_vector(p, keyg, key):
    # | firstsource
    if len(p) == 4:
        # create a pseudo p vector
        v = []
        v.append(p[0])
        # reappend the new line
        p[1][keyg][len(p[1][keyg]) - 1][key] += '\n'
        v.append(p[1])
        v.append(p[3])
        append_or_create(keyg, v)
        r = v[0]
    elif len(p) == 2:
        r = { keyg: [p[1]] }
    else:
        if keyg in p[1]:
            r = p[1]
            r[keyg].append(p[2])
        else:
            r = { keyg: [p[1], p[2]] }
    return r
 
def p_sourcelines(p):
    '''sourcelines : sourcelines sourceline
                   | sourcelines firstsource
                   | sourcelines firstdiagram
                   | firstsource sourceline
                   | firstsource 
                   | sourcelines NEWLINE sourceline
                   | sourcelines NEWLINE firstsource'''
    p[0] = insert_source_diagram_vector(p, 'sourcelines', 'string')

def p_diagramlines(p):
    '''diagramlines : diagramlines sourceline
                    | diagramlines firstsource
                    | diagramlines firstdiagram
                    | firstdiagram sourceline
                    | firstdiagram
                    | diagramlines NEWLINE sourceline
                    | diagramlines NEWLINE firstdiagram'''
    p[0] = insert_source_diagram_vector(p, 'diagramlines', 'string')

def p_textlines(p):
    '''textlines : textlines textline
                 | textline'''
    append_or_create('textlines', p)

def p_deflist(p):
    ''' deflist : textlines deflines '''
    p[0] = { 'deflist': { 'term': p[1], 'definition': p[2] } }

def p_deflistline(p):
    '''deflines : deflines firstdiagram
                 | firstdiagram'''
    append_or_create('textlines', p)

def p_firstsource(p):
    '''firstsource : FIRSTSOURCE NEWLINE
                   | TEXTLIST NEWLINE'''
    p[0] = { 'string': p[1] } 

def p_firstdiagram(p):
    '''firstdiagram : FIRSTDIAGRAM NEWLINE'''
    p[0] = { 'string': p[1] } 

def p_sourceline(p):
    '''sourceline : SOURCE NEWLINE
                  | TEXTLIST NEWLINE'''
    p[0] = { 'string': p[1]} 

def p_textline(p):
    '''textline : TEXT NEWLINE
                | TEXT2 NEWLINE'''
    p[0] = { 'string': p[1] }

def firstnospace (s):
    n = 0
    for c in s:
        if c == ' ':
            n += 1
        else:
            return n
    return None

def p_listline(p):
    '''listline : TEXTLIST NEWLINE'''
    # The number of spaces determines the indentation
    level = p[1].count(' ', 0, p[1].find('-'))
    p[0] = { 'level': level , 'string': p[1][(level + 2):] }

# the firstlistline is special because it determines wether the block
# is a list paragraph or not
# Todo: allow for various lines
def p_firstlist(p):
    '''firstlist : FIRSTLIST NEWLINE'''
    p[0] = { 'level': 0 , 'string': p[1][2:] }

def p_title(p):
    'title : textlines NEWLINE'
    # Copy directly the textlines as title is
    # prefixed by parent 'header', after removing any 'subtitle'
    para = p[1]
    if test_subtitle(para['textlines'][1]['string']):
        para['textlines'] = para['textlines'][:-1]
    p[0] = para

def p_header(p):
    'header : title textparas listlines NEWLINE'
    # The real rule (the resulting tree will have):
    # title paragraphs toc NEWLINE
    p[0] = { 'title': p[1], 'abstract': p[2], 'toc': p[3] }
    in_header = False

def p_body(p):
    'body : paragraphs'
    # In the tree there will not be a 'paragraphs'
    # branch, but several 'section' branches. As titles
    # have no special markup and have to be matched against
    # toc titles, no grammar can define them (better said,
    # i do not know how)
    p[0] = p[1]

def p_document(p):
    '''document : header body'''
    p[0] = { 'header': p[1], 'body': p[2]}
    document = p[0]
    # print str(p[0])

def p_error(t):
    if t:
        print "Syntax error at line %d, around '%s'" % (t.lineno, t.value)
    else:
        print "Syntax error at EOF"
    sys.exit(0)

# Detect any --- === ~~~~ that could have been added
def test_subtitle(s):
    for c in ['-', '=', '~']:
        test = c * len(s)
        if test == s:
            return True
    return False

def adjust_sections(doc):
    ''' Walk the parsed tree looking for paragraphs that are title
        sections and reorganize the tree with sections '''
    sections = []
    paras = doc['body']['paragraphs']
    nparas = len(paras)
    last = -1
    lastlevel = -1
    for i in range(nparas):
        para = paras[i]
        if para.has_key('textlines') == False:
            continue
        if len(para['textlines']) == 2:	  # A paragraph with 2 lines
            if test_subtitle(para['textlines'][1]['string']):
                para['textlines'] = para['textlines'][:-1]
            else:
                continue
        elif len(para['textlines']) != 1:
            continue

        title_candidate = para['textlines'][0]['string']

        toc_string, level = in_toclist(title_candidate, doc)
        if toc_string == None:
            continue
        para['textlines'][0]['string'] = toc_string
        if last == -1:
            last = i
            lastlevel = level
            continue
        lastpara = paras[last]
        section = { 'title': lastpara['textlines'][0], 'level': lastlevel, 'paragraphs' : paras[last+1:i]}
        sections.append(section)
        last = i
        lastlevel = level

    # Last section
    lastpara = paras[last]
    section = { 'title': lastpara['textlines'][0] , 'level': lastlevel, 'paragraphs' : paras[last+1:nparas] }
    sections.append(section)
    
    # From now on paragraphs are organiced into sectors, so delete
    # old paragraphs array
    doc['body']['sections'] = sections
    del doc['body']['paragraphs']

# Append with backslashes
def append(base, new):
    bl = len(base)
    if bl < 2:
        return base + new + '\n'
    if base[(bl-2):] == '\\\n':
        n = firstnospace(new)
        if n == None:
            return base
        return base[:-2] + new[n:] + '\n'
    else:
        return base + new + '\n'

def preprocess(s):
    # line continuations are very tricky to handle inside the grammar,
    # so i took this 'preprocessor' out of the sleeve :>
    o = ''
    e = ''
    n = 0
    inlist = False

    for line in s.split('\n'):
        l = len(line)
        if l == 0:
            if inlist:
                inlist = False
            o = append(o, '')

        if len(line) != 0 and line[0] != ' ':
            l = len(line)
            if len(line) > 72:
                e += 'Line ' + str(n) 
                e += ':In regular paragraphs (no source/diagrams) no'
                e += 'lines bigger than 72 chars are allowed. Use \\'
                return None, e
            else:
                if line[0] == '-':
                    inlist = True
            o = append(o, line)
        if len(line) != 0 and line[0] == ' ':
                if inlist:
                    n = firstnospace(line)
                    if line[n] == '-':
                        o = append(o, line)
                    else:
                        if o[-2:] == '\\\n':
                            o = append(o, line[n:])
                        else:
                            o = append(o[:-1], ' ' + line[n:])
                else:
                    insource = True
                    # No backslash substitution for source
                    # code
                    o += line + '\n'
        n += 1

    # Remove extra \n's at eol
    while o[-1:] == '\n':
        o = o[:-1]
    o += '\n\n'
    return o, None

def extract_links_lines(links, lines):
    ''' Extract links from a group of lines. If the group
        of lines only have links True is returned along with
        the modified array
    '''
    n = []
    links_only = True
    for l in lines:
        r = extract_links(l['string'])
        if r['links'] != {}:
            links.update(r['links'])
        else:
            links_only = False
        if r['references'] != []:
            l['string'] = r['modified_string']
            if l['string'] != '':
                links_only = False
            l['references'] = r['references']
        n.append(l)
    return n, links_only

def extract_links_paragraph(links, para):
    '''
    extract links from listlines and textlines paragraphs only
    (no source/diagrams etc)
    '''
    if para.has_key('listlines'):
        para['listlines'], o = extract_links_lines(links, para['listlines'])
        if o == True:
            del para['listlines']
    elif para.has_key('textlines'):
        para['textlines'], o = extract_links_lines(links, para['textlines'])

def process_links(doc):
    links = {}
    ''' Extract links from source and put them in their tree '''
    for para in doc['header']['abstract']['abstract']:
        extract_links_paragraph(links, para)
    for section in doc['body']['sections']:
        for para in section['paragraphs']:
            extract_links_paragraph(links, para)
    doc['links'] = links

def extract_attributes(doc):
    pass

def parse(s):
    ''' Return the syntax tree for a preloaded string '''
    l = lex.lex(optimize=1, debug=0)
    yacc.yacc(optimize=1, debug=0)
    p, e = preprocess(s)
    if p == None:
        print 'ERROR: ' + e
        return None
    doc = yacc.parse(p)
    # First, adjust 'named' sections from the template
    adjust_sections(doc)
    # Extract links from source and put them in their tree
    process_links(doc)
    # Identify text attributes
    extract_attributes(doc)
    # Parsetab by default is generated on the current directory
    # This is not desirable at all (the directory can be read-only
    # and a program should not write spureous files on its cwd)
    # But, until i figure out a better way to put the parsetab
    # in the installation dir (wich would be the best) just remove
    # it
    if os.path.exists('parsetab.py'):
        os.remove('parsetab.py')
    if os.path.exists('lextab.py'):
        os.remove('lextab.py')
    return doc, None

if __name__ == "__main__":
    import doctest
    doctest.testmod()

